"""lmdirect package for connecting to the local La Marzocco API."""
import asyncio
import logging

from lmdirect.const import (
    DISABLED,
    ENABLED,
    MACHINE_NAME,
    MODEL_NAME,
    SERIAL_NUMBER,
)

from .connection import Connection
from .msgs import (
    AUTO,
    AUTO_BITFIELD,
    AUTO_BITFIELD_MAP,
    DAYS,
    DOSE,
    DOSE_HOT_WATER,
    ENABLE_PREBREWING,
    ENABLE_PREINFUSION,
    FIRMWARE_VER,
    GLOBAL,
    HOUR,
    MIN,
    MODEL_GS3_AV,
    MODEL_LM,
    MSGS,
    OFF,
    ON,
    POWER,
    PREBREW_FLAG,
    PREBREWING,
    PREINFUSION,
    TOFF,
    TON,
    TSET_COFFEE,
    TSET_STEAM,
    STEAM_BOILER_ENABLE,
    TYPE_AUTO_ON_OFF,
    TYPE_COFFEE_TEMP,
    TYPE_MAIN,
    TYPE_PREBREW,
    TYPE_STEAM_TEMP,
    TYPE_PREINFUSION,
    TYPE_STEAM_BOILER_ENABLE,
    Msg,
)

_LOGGER = logging.getLogger(__name__)


class LMDirect(Connection):
    """Represents and provides an interface to the local API to a La Marzocco espresso machine"""

    def __init__(self, machine_info):
        super().__init__(machine_info)

        """Create locks to ensure correct previous status is retrieved for each service call"""
        self._locks = {
            x: asyncio.Lock()
            for x in [
                Msg.SET_POWER,
                Msg.SET_AUTO_ON_OFF_ENABLE,
                Msg.SET_AUTO_ON_OFF_TIMES,
                Msg.SET_DOSE,
                Msg.SET_DOSE_HOT_WATER,
                Msg.SET_PREBREW_TIMES,
                Msg.SET_COFFEE_TEMP,
                Msg.SET_STEAM_TEMP,
                Msg.SET_PREBREWING_ENABLE,
                Msg.SET_PREINFUSION_TIME,
                Msg.SET_STEAM_BOILER_ENABLE,
            ]
        }

    @property
    def current_status(self):
        """Return a dict of all the properties that have been received."""
        return self._current_status

    @property
    def machine_name(self):
        """Return the name of the machine."""
        return self._machine_info[MACHINE_NAME]

    @property
    def serial_number(self):
        """Return serial number."""
        return self._machine_info[SERIAL_NUMBER]

    @property
    def model_name(self):
        """Return model name."""
        return self._machine_info[MODEL_NAME]

    @property
    def firmware_version(self):
        """Return firmware version."""
        return self._current_status.get(FIRMWARE_VER, "Unknown")

    def register_callback(self, callback):
        """Register callback for updates."""
        if callable(callback):
            self._callback_list.append(callback)

    def register_raw_callback(self, msg, callback, **kwargs):
        """Register a callback for the raw response to a command."""
        if callable(callback):
            self._raw_callback_list.append((msg, callback))

    def deregister_raw_callback(self, key):
        """Register a callback for the raw response to a command."""

        """The key is a tuple comprised of (msg, callback)."""
        if key in self._raw_callback_list:
            self._raw_callback_list.remove(key)

    async def connect(self):
        """Connect to the machine."""
        return await self._connect()

    async def request_status(self):
        """Request new data."""
        msgs = [
            Msg.GET_STATUS,
            Msg.GET_CONFIG,
            Msg.GET_AUTO_ON_OFF_TIMES,
            Msg.GET_DRINK_STATS,
            Msg.GET_USAGE_STATS,
            Msg.GET_FRONT_DISPLAY,
            Msg.GET_PREINFUSION_TIMES,
            Msg.GET_FACTORY_CONFIG
        ]

        _LOGGER.debug("Requesting status")
        await asyncio.gather(*[self._send_msg(msg) for msg in msgs])

        """Also wait for current temp"""
        self._responses_waiting.append(MSGS[Msg.GET_TEMP_REPORT].msg)

    async def send_msg(self, msg_id, **kwargs):
        """Send a message to the machine."""
        await self._send_msg(msg_id, **kwargs)

    async def close(self):
        """Close the connection to the machine."""

        """Wait for the read task to exit"""
        if self._read_reaper_task:
            await asyncio.gather(self._read_reaper_task)
        await self._close()

    def call_callbacks(self, **kwargs):
        """Call all callbacks to refresh data for listeners."""
        self._call_callbacks(**kwargs)

    """Utils"""

    def _convert_to_ascii(self, value, size):
        """Convert an integer value to ASCII-encoded hex."""
        return ("%0" + str(size * 2) + "X") % value

    def _findkey(self, find_value, dict):
        """Find a key from the value in a dict."""
        return next(
            (key for key, value in dict.items() if value == find_value),
            None,
        )

    """Services"""

    async def set_power(self, power):
        """Send power on or power off commands."""
        async with self._locks[Msg.SET_POWER]:
            power_value = 1 if power else 0
            value = self._convert_to_ascii(power_value, size=1)
            await self._send_msg(Msg.SET_POWER, data=value)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[POWER] = power_value

            self._call_callbacks(entity_type=TYPE_MAIN)

    async def set_auto_on_off_enable(self, day_of_week=None, enable=None):
        """Configure auto on/off."""

        async with self._locks[Msg.SET_AUTO_ON_OFF_ENABLE]:
            if None in [day_of_week, enable]:
                raise InvalidInput(f"Some parameters invalid {day_of_week=} {enable=}")

            """We need the existing register value, so fail if we don't have it yet."""
            if AUTO_BITFIELD not in self._current_status:
                """Kick off a query so that we'll have the data later."""
                await self._send_msg(Msg.GET_AUTO_ON_OFF_TIMES)
                raise NotReady(f"Query not completed yet")

            """Extract value for this field."""
            bitfield = self._current_status[AUTO_BITFIELD]
            bitmask = 0x01 << self._findkey((day_of_week, AUTO), AUTO_BITFIELD_MAP)
            bitfield = bitfield | bitmask if enable else bitfield & ~bitmask
            buf_to_send = self._convert_to_ascii(bitfield, 1)

            _LOGGER.debug(
                f"set_on_off_enable: {buf_to_send=}, {MSGS[Msg.SET_AUTO_ON_OFF_ENABLE].msg=}"
            )
            await self._send_msg(Msg.SET_AUTO_ON_OFF_ENABLE, data=buf_to_send)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[day_of_week] = ENABLED if enable else DISABLED
                state[AUTO_BITFIELD] = bitfield

            self._call_callbacks(entity_type=TYPE_AUTO_ON_OFF)

    async def set_auto_on_off_global(self, value):
        """Set global auto on/off."""
        await self.set_auto_on_off_enable(GLOBAL, value)

    async def set_auto_on_off_times(
        self,
        day_of_week=None,
        hour_on=None,
        minute_on=None,
        hour_off=None,
        minute_off=None,
    ):
        """Configure auto on/off hours."""
        async with self._locks[Msg.SET_AUTO_ON_OFF_TIMES]:
            if None in [day_of_week, hour_on, minute_on, hour_off, minute_off]:
                raise InvalidInput(
                    f"Some parameters invalid {day_of_week=} {hour_on=} {minute_on=} {hour_off=} {minute_off=}"
                )

            isinstance(hour_on, str) and (hour_on := int(hour_on))
            isinstance(minute_on, str) and (minute_on := int(minute_on))

            isinstance(hour_off, str) and (hour_off := int(hour_off))
            isinstance(minute_off, str) and (minute_off := int(minute_off))

            """Validate input."""
            if not (
                0 <= hour_on <= 23
                and 0 <= minute_on <= 59
                and 0 <= hour_off <= 23
                and 0 <= minute_off <= 59
                and day_of_week in DAYS
            ):
                raise InvalidInput(
                    f"set_auto_on_off_times: Invalid values {day_of_week=} {hour_on=} {minute_on=} {hour_off=} {minute_off=}"
                )

            """Update hours."""
            data = self._convert_to_ascii(hour_on, size=1) + self._convert_to_ascii(
                hour_off, size=1
            )
            address_base = self._convert_to_ascii(
                Msg.AUTO_ON_OFF_HOUR_BASE + (DAYS.index(day_of_week) * 2), size=1
            )
            _LOGGER.debug(
                f"set_on_off_times: {data=}, {address_base=}, {MSGS[Msg.SET_AUTO_ON_OFF_TIMES].msg=}"
            )
            await self._send_msg(Msg.SET_AUTO_ON_OFF_TIMES, base=address_base, data=data)

            """Update minutes."""
            data = self._convert_to_ascii(minute_on, size=1) + self._convert_to_ascii(
                minute_off, size=1
            )
            address_base = self._convert_to_ascii(
                Msg.AUTO_ON_OFF_MIN_BASE + (DAYS.index(day_of_week) * 2), size=1
            )
            _LOGGER.debug(
                f"set_on_off_times: {data=}, {address_base=}, {MSGS[Msg.SET_AUTO_ON_OFF_TIMES].msg=}"
            )
            await self._send_msg(Msg.SET_AUTO_ON_OFF_TIMES, base=address_base, data=data)

        """Update the stored values to immediately reflect the change"""
        for state in [self._temp_state, self._current_status]:
            state[self._get_key((day_of_week, ON, HOUR))] = hour_on
            state[self._get_key((day_of_week, ON, MIN))] = minute_on
            state[self._get_key((day_of_week, OFF, HOUR))] = hour_off
            state[self._get_key((day_of_week, OFF, MIN))] = minute_off

        self.calculate_auto_sched_times(day_of_week)

        self._call_callbacks(entity_type=TYPE_MAIN)

    async def set_dose(self, key=None, pulses=None):
        """Set the coffee dose in pulses (~0.5ml)."""

        async with self._locks[Msg.SET_DOSE]:
            if None in [key, pulses]:
                raise InvalidInput(
                    f"set_dose: Some parameters not specified {key=} {pulses=}"
                )

            isinstance(pulses, str) and (pulses := int(pulses))
            isinstance(key, str) and (key := int(key))

            """Validate input."""
            if not (1 <= pulses <= 1000 and 1 <= key <= 5):
                raise InvalidInput(f"set_dose: Invalid values {pulses=} {key=}")

            data = self._convert_to_ascii(pulses, size=2)
            key = self._convert_to_ascii(Msg.DOSE_KEY_BASE + (key - 1) * 2, size=1)
            await self._send_msg(Msg.SET_DOSE, base=key, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[self._get_key((DOSE, f"k{key}"))] = pulses

            self._call_callbacks(entity_type=TYPE_MAIN)

    async def set_dose_hot_water(self, seconds=None):
        """Set the hot water dose in seconds."""

        async with self._locks[Msg.SET_DOSE_HOT_WATER]:
            if seconds is None:
                raise InvalidInput("set_dose_hot_water: Seconds not specified")

            isinstance(seconds, str) and (seconds := int(seconds))

            """Validate input."""
            if not (1 <= seconds <= 30):
                raise InvalidInput(f"Invalid values {seconds=}")

            data = self._convert_to_ascii(seconds, size=1)
            await self._send_msg(Msg.SET_DOSE_HOT_WATER, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[DOSE_HOT_WATER] = seconds

            self._call_callbacks(entity_type=TYPE_MAIN)

    async def set_prebrew_times(self, key=None, seconds_on=None, seconds_off=None):
        """Set prebrew on/off times in seconds."""

        async with self._locks[Msg.SET_PREBREW_TIMES]:
            if None in [key, seconds_on, seconds_off]:
                raise InvalidInput(
                    "set_prebrew_times: Some parameters invalid {key=} {seconds_on=} {seconds_off=}"
                )

            seconds_on, seconds_off = [float(x) for x in [seconds_on, seconds_off]]
            isinstance(key, str) and (key := int(key))

            """Validate input."""
            if not (
                0 <= seconds_on <= 5.9 and 0 <= seconds_off <= 5.9 and 1 <= key <= 4
            ):
                raise InvalidInput(f"Invalid values {seconds_on=} {seconds_off=}")

            if not (
                (self.model_name == MODEL_GS3_AV and 1 <= key <= 4)
                or (self.model_name == MODEL_LM and key == 1)
            ):
                raise InvalidInput(f"Invalid values {key=}")

            """Set "on" time."""
            key_on = self._convert_to_ascii(Msg.PREBREW_ON_BASE + (key - 1), size=1)
            data = self._convert_to_ascii(int(seconds_on * 10), size=1)
            await self._send_msg(Msg.SET_PREBREW_TIMES, base=key_on, data=data)

            """Set "off" time."""
            key_off = self._convert_to_ascii(Msg.PREBREW_OFF_BASE + (key - 1), size=1)
            data = self._convert_to_ascii(int(seconds_off * 10), size=1)
            await self._send_msg(Msg.SET_PREBREW_TIMES, base=key_off, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[self._get_key((PREBREWING, TON, f"k{key}"))] = seconds_on
                state[self._get_key((PREBREWING, TOFF, f"k{key}"))] = seconds_off

            self._call_callbacks(entity_type=TYPE_PREBREW)

    async def set_preinfusion_time(self, key=None, seconds=None):
        """Set preinfusion times in seconds."""

        async with self._locks[Msg.SET_PREINFUSION_TIME]:
            if None in [key, seconds]:
                raise InvalidInput(
                    "set_preinfusion_time: Some parameters invalid {key=} {seconds=}"
                )

            seconds = float(seconds)
            isinstance(key, str) and (key := int(key))

            """Validate input."""
            if not (
                0 <= seconds <= 5.9 and 1 <= key <= 4
            ):
                raise InvalidInput(f"Invalid values {seconds=}")

            if not (
                (self.model_name == MODEL_GS3_AV and 1 <= key <= 4)
                or (self.model_name == MODEL_LM and key == 1)
            ):
                raise InvalidInput(f"Invalid values {key=}")

            """Set "on" time."""
            key = self._convert_to_ascii(Msg.PREINFUSION_BASE + (key - 1), size=1)
            data = self._convert_to_ascii(int(seconds * 10), size=1)
            await self._send_msg(Msg.SET_PREINFUSION_TIME, base=key, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[self._get_key((PREINFUSION, f"k{key}"))] = seconds

            self._call_callbacks(entity_type=TYPE_PREBREW)

    async def set_coffee_temp(self, temp=None):
        """Set the coffee boiler temp in Celcius."""

        async with self._locks[Msg.SET_COFFEE_TEMP]:
            if temp is None:
                raise InvalidInput("set_coffee__temp: Temperature not specified")

            isinstance(temp, str) and (temp := float(temp))
            temp = round(temp, 1)

            data = self._convert_to_ascii(int(temp * 10), size=2)
            await self._send_msg(Msg.SET_COFFEE_TEMP, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[TSET_COFFEE] = temp

            self._call_callbacks(entity_type=TYPE_COFFEE_TEMP)

    async def set_steam_temp(self, temp=None):
        """Set the steam boiler temp in Celcius."""

        async with self._locks[Msg.SET_STEAM_TEMP]:
            if temp is None:
                raise InvalidInput("set_steam_temp: Temperature not specified")

            isinstance(temp, str) and (temp := float(temp))
            temp = round(temp, 1)

            data = self._convert_to_ascii(int(temp * 10), size=2)
            await self._send_msg(Msg.SET_STEAM_TEMP, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[TSET_STEAM] = temp

            self._call_callbacks(entity_type=TYPE_STEAM_TEMP)

    async def set_prebrewing_enable(self, enable):
        """Turn prebrewing on or off."""

        async with self._locks[Msg.SET_PREBREWING_ENABLE]:
            value = 1 if enable else 0
            data = self._convert_to_ascii(value, size=1)

            await self._send_msg(Msg.SET_PREBREWING_ENABLE, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[PREBREW_FLAG] = value
                state[ENABLE_PREBREWING] = (value == 1)
                state[ENABLE_PREINFUSION] = (value == 2)

            self._call_callbacks(entity_type=TYPE_PREBREW)
            self._call_callbacks(entity_type=TYPE_PREINFUSION)

    async def set_preinfusion_enable(self, enable):
        """Turn preinfusion on or off."""

        """Preinfusion is essentially a variant of prebrewing."""
        async with self._locks[Msg.SET_PREBREWING_ENABLE]:
            value = 2 if enable else 0
            data = self._convert_to_ascii(value, size=1)

            await self._send_msg(Msg.SET_PREBREWING_ENABLE, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[PREBREW_FLAG] = value
                state[ENABLE_PREINFUSION] = (value == 2)
                state[ENABLE_PREBREWING] = (value == 1)

            self._call_callbacks(entity_type=TYPE_PREINFUSION)
            self._call_callbacks(entity_type=TYPE_PREBREW)

    async def set_steam_boiler_enable(self, enable):
        """Enable or disable the steam boiler."""

        async with self._locks[Msg.SET_STEAM_BOILER_ENABLE]:
            value = 0x81 if enable else 0x01
            data = self._convert_to_ascii(value, size=1)

            await self._send_msg(Msg.SET_STEAM_BOILER_ENABLE, data=data)

            """Update the stored values to immediately reflect the change"""
            for state in [self._temp_state, self._current_status]:
                state[STEAM_BOILER_ENABLE] = enable

            self._call_callbacks(entity_type=TYPE_STEAM_BOILER_ENABLE)

    async def set_start_backflush(self):
        """Initiate a backflush cycle."""
        data = self._convert_to_ascii(0x82, size=1)

        await self._send_msg(Msg.SET_START_BACKFLUSH, data=data)

class InvalidInput(Exception):
    """Error to indicate that invalid parameters were provided."""

    def __init__(self, msg):
        _LOGGER.exception(msg)
        super().__init__(msg)


class NotReady(Exception):
    """Error to indicate that data hasn't been read yet that the service call relies on."""

    def __init__(self, msg):
        _LOGGER.exception(msg)
        super().__init__(msg)



"""lmdirect connection class."""
import asyncio
import logging
from datetime import datetime, timedelta
from functools import partial

from authlib.integrations.base_client.errors import OAuthError
from authlib.integrations.httpx_client import AsyncOAuth2Client

from .aescipher import AESCipher
from .const import *
from .msgs import (
    AUTO_BITFIELD,
    AUTO_BITFIELD_MAP,
    AUTO_SCHED_MAP,
    CURRENT_PULSE_COUNT,
    DAYS_SINCE_BUILT,
    DIVIDE_KEYS,
    DRINK_OFFSET_MAP,
    ENABLE_PREBREWING,
    ENABLE_PREINFUSION,
    FIRMWARE_VER,
    FRONT_PANEL_DISPLAY,
    GATEWAY_DRINK_MAP,
    HEATING_STATE,
    HEATING_VALUES,
    HOUR,
    KEY_ACTIVE,
    MIN,
    MSGS,
    OFF,
    ON,
    PREBREW_FLAG,
    SERIAL_NUMBERS,
    STEAM_BOILER_ENABLE,
    TEMP_COFFEE,
    TIME,
    TOTAL_COFFEE,
    TOTAL_COFFEE_ACTIVATIONS,
    TOTAL_FLUSHING,
    UNIT_FAHRENHEIT,
    UPDATE_AVAILABLE,
    BREW_GROUP_OFFSET,
    T_UNIT,
    FACTORY_OFFSET,
    Elem,
    Msg,
)

_LOGGER = logging.getLogger(__name__)


class Connection:
    def __init__(self, machine_info):
        """Init LMDirect."""
        self._reader = None
        self._writer = None
        self._read_response_task = None
        self._read_reaper_task = None
        self._current_status = {}
        self._responses_waiting = []
        self._run = True
        self._callback_list = []
        self._raw_callback_list = []
        self._cipher = None
        self._machine_info = machine_info
        self._start_time = None
        self._connected = False
        self._lock = asyncio.Lock()
        self._first_time = True
        self._update_available = None
        self._initialized_machine_info = False

        """Maintain temporary states for device states that take a while to update"""
        self._temp_state = {}

    def _get_key(self, k):
        """Construct tag name if needed."""
        if isinstance(k, tuple):
            k = "_".join(k)
        return k

    async def retrieve_machine_info(self, machine_info):
        """Retrieve the machine info from the cloud APIs."""
        _LOGGER.debug(f"Retrieving machine info")

        async with AsyncOAuth2Client(
            client_id=machine_info[CLIENT_ID],
            client_secret=machine_info[CLIENT_SECRET],
            token_endpoint=TOKEN_URL,
        ) as client:

            headers = {
                "client_id": machine_info[CLIENT_ID],
                "client_secret": machine_info[CLIENT_SECRET],
            }

            try:
                await client.fetch_token(
                    url=TOKEN_URL,
                    username=machine_info[USERNAME],
                    password=machine_info[PASSWORD],
                    headers=headers,
                )
            except OAuthError as err:
                raise AuthFail("Authorization failure") from err

            """Only retrieve info if we're missing something."""
            if any(
                x not in machine_info
                for x in [KEY, SERIAL_NUMBER, MACHINE_NAME, MODEL_NAME]
            ):
                cust_info = await client.get(CUSTOMER_URL)
                if cust_info:
                    fleet = cust_info.json()["data"]["fleet"][0]
                    machine_info[KEY] = fleet["communicationKey"]
                    machine_info[SERIAL_NUMBER] = fleet["machine"]["serialNumber"]
                    machine_info[MACHINE_NAME] = fleet["name"]
                    machine_info[MODEL_NAME] = fleet["machine"]["model"]["name"]

            """Add the machine and model names to the dict so that they're avaialble for attributes"""
            self._current_status.update({MACHINE_NAME: machine_info[MACHINE_NAME]})
            self._current_status.update({MODEL_NAME: machine_info[MODEL_NAME]})

            if any(
                self._get_key(x) not in self._current_status
                for x in DRINK_OFFSET_MAP.values()
            ):
                drink_info = await client.get(
                    DRINK_COUNTER_URL.format(serial_number=machine_info[SERIAL_NUMBER])
                )
                if drink_info:
                    data = drink_info.json().get("data")
                    if data:
                        self._current_status.update(
                            {
                                self._get_key(GATEWAY_DRINK_MAP[x["coffeeType"]]): x["count"]
                                for x in data
                            }
                        )

            if UPDATE_AVAILABLE not in self._current_status:
                update_info = await client.get(UPDATE_URL)
                if update_info:
                    data = update_info.json().get("data")
                    self._update_available = "Yes" if data else "No"

        self._initialized_machine_info = True

        _LOGGER.debug(f"Finished machine info")
        return machine_info

    async def _connect(self):
        """Conmnect to espresso machine."""
        if self._connected:
            return self._machine_info

        _LOGGER.debug(f"Connecting")

        if not self._initialized_machine_info:
            try:
                self._machine_info = await self.retrieve_machine_info(
                    self._machine_info
                )
            except AuthFail as err:
                raise err
            except Exception as err:
                raise ConnectionFail(
                    f"Exception retrieving machine info: {err}"
                ) from err

        if not self._cipher:
            self._cipher = AESCipher(self._machine_info[KEY])

        """Connect to the machine."""
        try:
            self._reader, self._writer = await asyncio.wait_for(
                asyncio.open_connection(
                    self._machine_info[HOST], self._machine_info[PORT]
                ),
                timeout=3,
            )
        except asyncio.TimeoutError:
            _LOGGER.warning("Connection Timeout, skipping")
            return None

        except Exception as err:
            raise ConnectionFail(f"Cannot connect to machine: {err}") from err

        """Start listening for responses."""
        await self.start_read_task()

        self._connected = True
        _LOGGER.debug("Finished Connecting")

        return self._machine_info

    async def start_read_task(self):
        """Start listening for responses."""
        loop = asyncio.get_event_loop()
        self._read_response_task = loop.create_task(
            self.read_response_task(), name="Response Task"
        )

        """Reap the results and any any exceptions"""
        self._read_reaper_task = loop.create_task(
            self.read_reaper(), name="Read Reaper"
        )

    async def _close(self):
        """Close the connection to the machine."""
        if not self._connected:
            return

        if self._writer is not None:
            self._writer.close()

        if self._read_response_task:
            self._read_response_task.cancel()

        self._reader = self._writer = None
        self._connected = False
        _LOGGER.debug("Finished closing")

    async def read_reaper(self):
        _LOGGER.debug("Starting read reaper")
        try:
            await asyncio.gather(self._read_response_task)
        except Exception as err:
            _LOGGER.error(f"Exception in read_response_task: {err}")

        await self._close()
        self._read_response_task = None
        self._first_time = False

        _LOGGER.debug("Finished reaping read task")

    def _call_callbacks(self, **kwargs):
        """Call the callbacks."""
        if self._callback_list is not None:
            [
                elem(
                    current_status=self._current_status,
                    **kwargs,
                )
                for elem in self._callback_list
            ]

    async def read_response_task(self):
        """Start thread to receive responses."""
        BUFFER_SIZE = 1000
        handle = None

        _LOGGER.debug("Starting read task")

        if self._start_time is None:
            self._start_time = datetime.now()

        while self._run:
            encoded_data = await self._reader.readuntil(separator=b"%")

            if encoded_data is not None:
                loop = asyncio.get_event_loop()
                fn = partial(self._cipher.decrypt, encoded_data[1:-1])
                plaintext = await loop.run_in_executor(None, fn)
                if not plaintext:
                    continue

                await self.process_data(plaintext)

                if not self._first_time:
                    if handle:
                        handle.cancel()
                        handle = None

                    """Coalesce callbacks within a 5s window."""
                    handle = loop.call_later(5, self._call_callbacks)
                else:
                    self._call_callbacks()

                """Exit if we've been reading longer than 10s since the last command."""
                if datetime.now() > self._start_time + timedelta(seconds=10):
                    _LOGGER.debug(f"Exiting loop: {self._responses_waiting}")

                    """Flush the wait list."""
                    self._responses_waiting = []
                    break

    async def process_data(self, plaintext):
        """Process incoming packet."""

        """Separate the mesg from the data."""
        msg_type = plaintext[0]
        raw_data = plaintext[1:]
        msg = raw_data[:8]
        retval = True

        """Chop off the message and check byte."""
        data = raw_data[8:-2]
        finished = not len(self._responses_waiting)

        _LOGGER.debug(f"Message={msg}, Data={data}")

        msg_id = None

        if msg_type == Msg.WRITE:
            if Msg.RESPONSE_GOOD not in data:
                _LOGGER.error(f"Command Failed: {msg}: {data}")
                retval = False
            else:
                _LOGGER.debug(f"Command Succeeded: {msg}: {data}")
        else:
            """Find the matching item or returns None."""
            msg_id = next(
                (
                    x
                    for x in MSGS
                    if MSGS[x].msg == msg and MSGS[x].msg_type == msg_type
                ),
                None,
            )

            if msg_id is not None:
                cur_msg = MSGS[msg_id]

                """Notify any listeners for this message."""
                [
                    await x[1]((msg_id, x[1]), data)
                    for x in self._raw_callback_list
                    if MSGS[x[0]].msg == msg
                ]

                if cur_msg.map is not None:
                    await self._populate_items(data, cur_msg)
            else:
                _LOGGER.error(f"Unexpected response: {plaintext}")
                retval = False

        if msg in self._responses_waiting:
            self._responses_waiting.remove(msg)
            finished = not len(self._responses_waiting)
            if finished:
                _LOGGER.debug("Received all responses")

        return retval

    def calculate_auto_sched_times(self, key):
        time_on_key = self._get_key((key, ON, TIME))
        hour_on_key = self._get_key((key, ON, HOUR))
        min_on_key = self._get_key((key, ON, MIN))
        time_off_key = self._get_key((key, OFF, TIME))
        hour_off_key = self._get_key((key, OFF, HOUR))
        min_off_key = self._get_key((key, OFF, MIN))

        """Set human-readable "on" time"""
        self._current_status[
            time_on_key
        ] = f"{'%02d' % self._current_status[hour_on_key]}:{'%02d' % self._current_status[min_on_key]}"

        """Set human-readable "off" time"""
        self._current_status[
            time_off_key
        ] = f"{'%02d' % self._current_status[hour_off_key]}:{'%02d' % self._current_status[min_off_key]}"

    async def _populate_items(self, data, cur_msg):
        def handle_cached_value(element, value):
            """See if we've stored a temporary value that may take a while to update on the machine"""
            if element in self._temp_state:
                if value == self._temp_state[element]:
                    """Value has updated, so remove the temp value"""
                    _LOGGER.debug(
                        f"Element {element} has updated to {value}, pop the cached value"
                    )
                    self._temp_state.pop(element, None)
                else:
                    """Value hasn't updated yet, so use the cached value"""
                    _LOGGER.debug(
                        f"Element {element} hasn't updated yet, so use the cached value {value}"
                    )
                    value = self._temp_state[element]

            return value

        """Process all the fields and populate shared dict."""
        map = cur_msg.map
        for elem in map:
            value = None

            """Don't decode a value if we just plan to calculate it"""
            if elem.index != CALCULATED_VALUE:
                """The strings are ASCII-encoded hex, so each value takes 2 bytes."""
                index = elem.index * 2
                size = elem.size * 2

                """Extract value for this field."""
                value = data[index : index + size]

                if elem.type == Elem.INT:
                    """Convert from ascii-encoded hex."""
                    value = int(value, 16)

            raw_key = map[elem]

            """Construct key name if needed."""
            key = self._get_key(raw_key)

            if any(x in key for x in DIVIDE_KEYS):
                value = value / 10
            elif key == FIRMWARE_VER:
                value = "%0.2f" % (value / 100)
            elif key in SERIAL_NUMBERS:
                value = "".join(
                    [chr(int(value[i : i + 2], 16)) for i in range(0, len(value), 2)]
                )
                """Chop off any trailing nulls."""
                value = value.partition("\0")[0]

            elif key == AUTO_BITFIELD:
                bitfield = value
                for item in AUTO_BITFIELD_MAP:
                    setting = ENABLED if bitfield & 0x01 else DISABLED
                    processed_key = self._get_key(AUTO_BITFIELD_MAP[item])
                    self._current_status[processed_key] = handle_cached_value(
                        processed_key, setting
                    )
                    bitfield = bitfield >> 1
            elif raw_key in DRINK_OFFSET_MAP:
                if key == TOTAL_FLUSHING:
                    value = (
                        self._current_status[TOTAL_COFFEE_ACTIVATIONS]
                        - self._current_status[TOTAL_COFFEE]
                    )
                offset_key = self._get_key(DRINK_OFFSET_MAP[raw_key])
                if key not in self._current_status:
                    """If we haven't seen the value before, calculate the offset."""
                    self._current_status.update(
                        {offset_key: value - self._current_status.get(offset_key, 0)}
                    )
                """Apply the offset to the value."""
                value = value - self._current_status.get(offset_key, 0)
            elif key == DAYS_SINCE_BUILT:
                """Convert hours to days."""
                value = round(value / 24)
            elif key == HEATING_STATE:
                value = [x for x in HEATING_VALUES if HEATING_VALUES[x] & value]
                """Don't add attribute and remove it if machine isn't currently running."""
                if not value:
                    self._current_status.pop(key, None)
                    continue
            elif key == BREW_GROUP_OFFSET:
                value = (value & 0xFF00) >> 8 | (value & 0x00FF) << 8
                units = self._current_status.get(T_UNIT, 0)
                factory_offset = self._current_status.get(FACTORY_OFFSET, 0)

                # The Linea Mini has no offset
                if factory_offset == 0:
                    value = 0
                else:
                    # convert to celcius if the machine is set to Fahrenheit
                    if units == UNIT_FAHRENHEIT:
                        value *= 200/360

                    value = 2 * round((value - 100) / 10, 1)
            elif key in [KEY_ACTIVE, CURRENT_PULSE_COUNT]:
                """Don't add attributes and remove them if machine isn't currently running."""
                if not value:
                    self._current_status.pop(key, None)
                    continue
            elif key == FRONT_PANEL_DISPLAY:
                value = (
                    bytes.fromhex(value)
                    .decode("latin-1")
                    .replace("\xdf", "\u00b0")  # Degree symbol
                    .replace(
                        "\xdb", "\u25A1"
                    )  # turn a block into an outline block (heating element off)
                    .replace(
                        "\xff", "\u25A0"
                    )  # turn \xff into a solid block (heating element on)
                )
            elif elem.index == CALCULATED_VALUE and key in AUTO_SCHED_MAP.values():
                self.calculate_auto_sched_times(key)
                continue
            elif key == STEAM_BOILER_ENABLE:
                value = (value & 0x01) == 1
            elif elem.index == CALCULATED_VALUE and key in [ENABLE_PREBREWING, ENABLE_PREINFUSION]:
                state = self._current_status.get(PREBREW_FLAG)
                value = state == (1 if key == ENABLE_PREBREWING else 2)

            if key == TEMP_COFFEE:
                # The offset is used to set the boiler temp to achieve the
                # user-set temp at the grouphead, so subtract to get the
                # group temp
                value = round(value - self._current_status.get(BREW_GROUP_OFFSET, 0), 1)

            self._current_status[key] = handle_cached_value(key, value)

    async def _send_msg(self, msg_id, data=None, base=None):
        """Send command to the espresso machine."""
        msg = MSGS[msg_id]

        _LOGGER.debug(f"Sending {msg.msg} with {data} {base}")
        await self._send_raw_msg(msg.msg, msg.msg_type, data, base)

    async def _send_raw_msg(self, msg, msg_type, data=None, base=None):
        def checksum(buffer):
            """Compute check byte."""
            buffer = bytes(buffer, "utf-8")
            return "%0.2X" % (sum(buffer) % 256)

        """Prevent race conditions - can be called from different tasks."""
        async with self._lock:
            """Connect if we don't have an active connection."""
            result = await self._connect()

            if not result:
                raise ConnectionFail("Connection failed.")

            if not self._writer:
                raise ConnectionFail(f"self._writer={self._writer}")

            """If a key was provided, replace the second byte of the message."""
            msg_to_send = msg if not base else msg[:2] + base + msg[4:]

            plaintext = msg_type + msg_to_send

            if data is not None:
                plaintext += data

            """Add the check byte."""
            plaintext += checksum(plaintext)

            loop = asyncio.get_event_loop()
            fn = partial(self._cipher.encrypt, plaintext)
            ciphertext = (
                "@" + (await loop.run_in_executor(None, fn)).decode("utf-8") + "%"
            )

            self._writer.write(bytes(ciphertext, "utf-8"))
            await self._writer.drain()

            """Remember that we're waiting for a response."""
            self._responses_waiting.append(msg_to_send)

            """Note when the command was sent."""
            self._start_time = datetime.now()


class AuthFail(Exception):
    """Error to indicate there is invalid auth info."""

    def __init__(self, msg):
        super().__init__(msg)


class ConnectionFail(Exception):
    """Error to indicate there is no connection."""

    def __init__(self, msg):
        super().__init__(msg)



"""Local API commands for La Marzocco espresso machines."""

"""All models currently Wifi-enabled."""
MODEL_GS3_AV = "GS3 AV"
MODEL_GS3_MP = "GS3 MP"
MODEL_LM = "Linea Mini"

# Tags

from lmdirect.const import (
    CALCULATED_VALUE,
    CONTINUOUS_OFFSET,
    FLUSHING_OFFSET,
)

OFFSET = "offset"

DATE_RECEIVED = "date_received"
POWER = "power"
POWER_MYSTERY = "power_mystery"
TEMP_COFFEE = "coffee_temp"
TEMP_STEAM = "steam_temp"
TSET_COFFEE = "coffee_set_temp"
TSET_STEAM = "steam_set_temp"

DOSE = "dose"

DOSE_HOT_WATER = "dose_hot_water"
PREBREW_FLAG = "prebrew_flag"
ENABLE_PREBREWING = "enable_prebrewing"
ENABLE_PREINFUSION = "enable_preinfusion"

PREBREWING = "prebrewing"
PREINFUSION = "preinfusion"
TON = "ton"
TOFF = "toff"

DRINKS = "drinks"

CONTINUOUS = "continuous"
TOTAL_COFFEE = "total_coffee"
HOT_WATER = "hot_water"
DRINK_MYSTERY = "drink_mystery"
TOTAL_COFFEE_ACTIVATIONS = "total_coffee_activations"
HOT_WATER_2 = "hot_water_2"
DRINKS_HOT_WATER = "drinks_hot_water"
TOTAL_FLUSHING = "total_flushing"
MACHINE_NAME = "machine_name"
FRONT_PANEL_DISPLAY = "front_panel_display"
MYSTERY_VALUES = "mystery_values"
BREW_GROUP_OFFSET = "brew_group_offset"
FACTORY_OFFSET = "factory_offset"
PID_OFFSET = "pid_offset"
T_UNIT = "t_unit"
WATER_FILTER_LITERS = "water_filter_liters"

UNIT_CELCIUS = 0
UNIT_FAHRENHEIT = 1

VAL = "val"

MYSTERY_1 = "mystery_1"
MYSTERY_2 = "mystery_2"
STEAM_BOILER_ENABLE = "steam_boiler_enable"
HEATING_STATE = "heating_state"
STEAM_HEATER_ON = "steam_heater_on"
COFFEE_HEATER_ON = "coffee_heater_on"
HEATING_ON = "heating_on"
PUMP_ON = "pump_on"
BREW_SOLENOID_ON = "brew_solenoid_on"
HOT_WATER_SOLENOID_ON = "hot_water_solenoid_on"
BOILER_FILL_SOLENOID_ON = "boiler_fill_solenoid_on"

KEY_ACTIVE = "key_active"
CURRENT_PULSE_COUNT = "current_pulse_count"
WATER_RESERVOIR_CONTACT = "water_reservoir_contact"

COFFEE_HEATING_ELEMENT_HOURS = "coffee_heating_element_hours"
STEAM_HEATING_ELEMENT_HOURS = "steam_heating_element_hours"
MACHINE_RUNNING_SECONDS = "machine_running_seconds"
DAYS_SINCE_BUILT = "days_since_built"
PUMP_ON_SECONDS = "pump_on_seconds"
WATER_ON_SECONDS = "water_on_seconds"

FIRMWARE_VER = "firmware_ver"
MODULE_SER_NUM = "module_ser_num"

UPDATE_AVAILABLE = "update_available"

ON = "on"
OFF = "off"
HOUR = "hour"
MIN = "min"
AUTO = "auto"
TIME = "time"

GLOBAL = "global"
MON = "mon"
TUE = "tue"
WED = "wed"
THU = "thu"
FRI = "fri"
SAT = "sat"
SUN = "sun"

DAYS = [MON, TUE, WED, THU, FRI, SAT, SUN]

AUTO_BITFIELD = "auto_bitfield"

SECOND = "second"
MINUTE = "minute"
HOUR = "hour"
DAYOFWEEK = "dayofweek"
DAY = "day"
MONTH = "month"
YEAR = "month"

FLOW_KEY = "flow_key"
FLOW_RATE = "flow_rate"
FLOW_PULSES = "flow_pulses"
FLOW_SECONDS = "flow_seconds"

WRITE_RESULT = "write_result"
MACHINE_SER_NUM = "machine_ser_num"

"""Data Types"""
TYPE_MAIN = 1
TYPE_PREBREW = 2
TYPE_AUTO_ON_OFF = 3
TYPE_COFFEE_TEMP = 4
TYPE_STEAM_TEMP = 5
TYPE_DRINK_STATS = 6
TYPE_WATER_RESERVOIR_CONTACT = 7
TYPE_PREINFUSION = 8
TYPE_START_BACKFLUSH = 9
TYPE_STEAM_BOILER_ENABLE = 10

DIVIDE_KEYS = ["temp", "prebrewing_to", "preinfusion_k", "pid_offset"]
SERIAL_NUMBERS = [MACHINE_SER_NUM, MODULE_SER_NUM]

# Response Maps
class Elem:

    INT = 0
    STRING = 1

    def __init__(self, index, size=1, type=INT):
        self._index = index
        self._size = size
        self._type = type

    @property
    def index(self):
        return self._index

    @property
    def size(self):
        return self._size

    @property
    def type(self):
        return self._type


# R
# 40 1C 00 04: Message
# 03 BC: Coffee Temp (95.6C)
# 04 D8: Steam Temp (124.0C)
# B6: Check byte

TEMP_REPORT_MAP = {
    Elem(0, 2): TEMP_COFFEE,
    Elem(2, 2): TEMP_STEAM
}

# R
# 40 00 00 20: Message
# 01
# 78: Version (1.20)
# 02
# 53 6E XX XX XX XX XX XX XX XX XX XX: Module serial number (SnXXXXXXXXXX)
# 00: ?? sometimes 01
# 00 00 00 00 00 00 00
# 27: 01: Power (ON)
# 01 00 00 00
# 32: 02 96: Coffee Temp (66.2C)
# 34: 03 2A: Steam Temp (81.0C)
# 70: Check byte

STATUS_MAP = {
    Elem(1): FIRMWARE_VER,
    Elem(3, 12, type=Elem.STRING): MODULE_SER_NUM,
    Elem(15): POWER_MYSTERY,
    Elem(17): KEY_ACTIVE,
    Elem(19, 2): CURRENT_PULSE_COUNT,
    # Elem(23): POWER,
    Elem(24): WATER_RESERVOIR_CONTACT,
    Elem(27): HEATING_STATE,
    #Elem(28, 2): TEMP_COFFEE,
    #Elem(30, 2): TEMP_STEAM,
    Elem(32): MYSTERY_1,
    Elem(33): MYSTERY_2,
    Elem(34): STEAM_BOILER_ENABLE,
}

HEATING_VALUES = {
    HEATING_ON: 0x10,
    COFFEE_HEATER_ON: 0x20,
    STEAM_HEATER_ON: 0x40,
    PUMP_ON: 0x01,
    BREW_SOLENOID_ON: 0x02,
    BOILER_FILL_SOLENOID_ON: 0x04,  # ??
    HOT_WATER_SOLENOID_ON: 0x08,
}

# R
# 00 00 00 1F: Preamble
# 01 00 00 02 6E 31 39
# 11: 03 C7: Coffee Temp Set (96.7C)
# 13: 04 D9: Steam Temp Set (124.1C)
# 15: 00: Prebrewing ON (off)
# 16: 00: B1 Prebrewing ON
# 17: 00: B2 Prebrewing ON
# 18: 00: B3 Prebrewing ON
# 19: 0A: B4 Prebrewing ON (1.0s)
# 20: 00: B1 Prebrewing Off
# 21: 00: B2 Prebrewing Off
# 22: 00: B3 Prebrewing Off
# 23: 00: B4 Prebrewing Off
# 24: 0078: Dose B1
# 26: 0076: Dost B2
# 28: 0078: Dose B3
# 30: 0085: Dose B4
# 32: 03E8: Dose B5
# 34: 08: Seconds hot water
# 35: Check byte

CONFIG_MAP = {
    Elem(0): POWER,
    Elem(7, 2): TSET_COFFEE,
    Elem(9, 2): TSET_STEAM,
    Elem(11): PREBREW_FLAG,
    Elem(CALCULATED_VALUE): ENABLE_PREBREWING,
    Elem(12): (PREBREWING, TON, "k1"),
    Elem(13): (PREBREWING, TON, "k2"),
    Elem(14): (PREBREWING, TON, "k3"),
    Elem(15): (PREBREWING, TON, "k4"),
    Elem(16): (PREBREWING, TOFF, "k1"),
    Elem(17): (PREBREWING, TOFF, "k2"),
    Elem(18): (PREBREWING, TOFF, "k3"),
    Elem(19): (PREBREWING, TOFF, "k4"),
    Elem(20, 2): (DOSE, "k1"),
    Elem(22, 2): (DOSE, "k2"),
    Elem(24, 2): (DOSE, "k3"),
    Elem(26, 2): (DOSE, "k4"),
    Elem(28, 2): (DOSE, "k5"),
    Elem(30): DOSE_HOT_WATER,
}

# Response to R 03 10 00 1D EB

# R
# 03 10 00 1D: Message
# 4: FE: Bitfield for auto on/off, where upper bits are the days bit 0 is auto on/off (disabled)
# 5: 06: Sunday on hour
# 6: 11: Sunday off hour (5pm)
# 7: 06: Monday on hour
# 8: 11: Monday off hour (5pm)
# 9: 06: Tuesday on hour
# 10: 11: Tuesday off hour (5pm)
# 11: 06: Wednesday on hour
# 12: 11: Wednesday off hour (5pm)
# 13: 06: Thursday on hour
# 14: 11: Thursday off hour (5pm)
# 15: 06: Friday on hour
# 16: 11: Friday off hour (5pm)
# 17: 06: Saturday on hour
# 18: 11: Saturday off hour (5pm)
# 00 00 00 00 00 00 00 00 00 00 00 00 00 00
# 2E: Check byte

AUTO_SCHED_MAP = {
    Elem(0): AUTO_BITFIELD,
    Elem(1): (MON, ON, HOUR),
    Elem(2): (MON, OFF, HOUR),
    Elem(3): (TUE, ON, HOUR),
    Elem(4): (TUE, OFF, HOUR),
    Elem(5): (WED, ON, HOUR),
    Elem(6): (WED, OFF, HOUR),
    Elem(7): (THU, ON, HOUR),
    Elem(8): (THU, OFF, HOUR),
    Elem(9): (FRI, ON, HOUR),
    Elem(10): (FRI, OFF, HOUR),
    Elem(11): (SAT, ON, HOUR),
    Elem(12): (SAT, OFF, HOUR),
    Elem(13): (SUN, ON, HOUR),
    Elem(14): (SUN, OFF, HOUR),
    Elem(15): (MON, ON, MIN),
    Elem(16): (MON, OFF, MIN),
    Elem(17): (TUE, ON, MIN),
    Elem(18): (TUE, OFF, MIN),
    Elem(19): (WED, ON, MIN),
    Elem(20): (WED, OFF, MIN),
    Elem(21): (THU, ON, MIN),
    Elem(22): (THU, OFF, MIN),
    Elem(23): (FRI, ON, MIN),
    Elem(24): (FRI, OFF, MIN),
    Elem(25): (SAT, ON, MIN),
    Elem(26): (SAT, OFF, MIN),
    Elem(27): (SUN, ON, MIN),
    Elem(28): (SUN, OFF, MIN),
    Elem(CALCULATED_VALUE): MON,
    Elem(CALCULATED_VALUE): TUE,
    Elem(CALCULATED_VALUE): WED,
    Elem(CALCULATED_VALUE): THU,
    Elem(CALCULATED_VALUE): FRI,
    Elem(CALCULATED_VALUE): SAT,
    Elem(CALCULATED_VALUE): SUN,
}

AUTO_ENABLE_MAP = {
    Elem(0): AUTO_BITFIELD,
}

AUTO_BITFIELD_MAP = {
    0: (GLOBAL, AUTO),
    1: (MON, AUTO),
    2: (TUE, AUTO),
    3: (WED, AUTO),
    4: (THU, AUTO),
    5: (FRI, AUTO),
    6: (SAT, AUTO),
    7: (SUN, AUTO),
}

# App: R 01 00 00 11 D5
# Machine:
# R
# 01 00 00 11: Read machine serial number
# 5A: Z (?)
# XX XX XX XX XX XX XX XX 00 00 00 00 00 00 00 00: Machine serial number
# 7D: Check byte

SER_NUM_MAP = {
    Elem(1, 16, type=Elem.STRING): MACHINE_SER_NUM,
}

# R
# 03000007
# 00: Second (00)
# 05: Minute (5)
# 0A: Hour (10)
# 02: Day of week (2)
# 16: Day (22)
# 0C: Month (12)
# 14: Year (20)
# B3: Check byte

DATETIME_MAP = {
    Elem(0): SECOND,
    Elem(1): MINUTE,
    Elem(2): HOUR,
    Elem(3): DAYOFWEEK,
    Elem(4): DAY,
    Elem(5): MONTH,
    Elem(6): YEAR,
}

# R
# 0020002C
# 0000014A: Key 1 (1 espresso) Completed Drinks
# 00000096: Key 2 (2 espressos) Completed Drinks
# 000001B0: Key 3 (1 coffee) Completed Drinks
# 00000021: Key 4 (2 coffees) Completed Drinks
# 00000563: Continuous Completed Drinks
# 00000914: Total Completed Coffee Drinks (total of above)
# 00000010: Hot water activations
# 0000000A: Increments every so often (?)
# 00000ADE: Total Coffee Drink Activations
# 00000010: Hot water activations
# 00000032: Completed Hot Water

DRINK_STATS_MAP = {
    Elem(0, 4): (DRINKS, "k1"),
    Elem(4, 4): (DRINKS, "k2"),
    Elem(8, 4): (DRINKS, "k3"),
    Elem(12, 4): (DRINKS, "k4"),
    Elem(16, 4): CONTINUOUS,
    Elem(20, 4): TOTAL_COFFEE,
    Elem(24, 4): HOT_WATER,
    Elem(28, 4): DRINK_MYSTERY,
    Elem(32, 4): TOTAL_COFFEE_ACTIVATIONS,
    Elem(36, 4): HOT_WATER_2,
    Elem(40, 4): DRINKS_HOT_WATER,
    Elem(CALCULATED_VALUE): TOTAL_FLUSHING,
}

GATEWAY_DRINK_MAP = {
    -1: FLUSHING_OFFSET,
    0: (DRINKS, "k1", OFFSET),
    1: (DRINKS, "k2", OFFSET),
    2: (DRINKS, "k3", OFFSET),
    3: (DRINKS, "k4", OFFSET),
    4: CONTINUOUS_OFFSET,
}

DRINK_OFFSET_MAP = {
    TOTAL_FLUSHING: FLUSHING_OFFSET,
    (DRINKS, "k1"): (DRINKS, "k1", OFFSET),
    (DRINKS, "k2"): (DRINKS, "k2", OFFSET),
    (DRINKS, "k3"): (DRINKS, "k3", OFFSET),
    (DRINKS, "k4"): (DRINKS, "k4", OFFSET),
    CONTINUOUS: CONTINUOUS_OFFSET,
}

# Z
# 60000016
# Snxxxxxxxxxx
# 04: Key
# 01: Rate (1 = high flow rate, 0 = low flow rate) - flips from 1 to 0 as flow slows
# 0009: Seconds (upper 12 bits are whole seconds, lower 4 bits are fractional)
# 000C: Pulses
# 03B6: Coffee boiler temp
# 04D4: Steam boiler temp
# DC: Check byte

WATER_FLOW_MAP = {
    Elem(12): FLOW_KEY,
    Elem(13): FLOW_RATE,
    Elem(14, 2): FLOW_SECONDS,
    Elem(16, 2): FLOW_PULSES,
    # Elem(18, 2): TEMP_COFFEE,
    # Elem(20, 2): TEMP_STEAM,
}

# R
# 00500018 - Message
# 000001FE - Coffee heating element minutes on
# 00000189 - Steam heating element minutes on
# 00009772 - Number of seconds that the machine has been running, either coffee or hot water
# 00001CA9 - Total hours of on-time
# 00009500 - Number of seconds that the pump has been running
# 00000075 - Number of seconds that hot water has been dispensed

USAGE_MAP = {
    Elem(0, 4): COFFEE_HEATING_ELEMENT_HOURS,
    Elem(4, 4): STEAM_HEATING_ELEMENT_HOURS,
    Elem(8, 4): MACHINE_RUNNING_SECONDS,
    Elem(12, 4): DAYS_SINCE_BUILT,
    Elem(16, 4): PUMP_ON_SECONDS,
    Elem(20, 4): WATER_ON_SECONDS,
}

# 3230322E39DF46: 202.9°F
# 20: Space
# DBDB: Boiler states
# 30393A32386100: 09:28a
# 42555A5A2020202020202020: BUZZ
# 543A3037: T:07


FRONT_DISPLAY_MAP = {
    Elem(0, 16, Elem.STRING): FRONT_PANEL_DISPLAY,
}

# 03
# E7
# 02
# D8
# 04
# F8
# 09
# 5B
# 02
# C2
# 03
# 04
# 08
# 9E
# 02
# 49
# 00
# F5

MYSTERY_MAP = {
    Elem(0): (VAL, 1),
    Elem(1): (VAL, 2),
    Elem(2): (VAL, 3),
    Elem(3): (VAL, 4),
    Elem(4): (VAL, 5),
    Elem(5): (VAL, 6),
    Elem(6): (VAL, 7),
    Elem(7): (VAL, 8),
    Elem(8): (VAL, 9),
    Elem(9): (VAL, 10),
    Elem(10): (VAL, 11),
    Elem(11): (VAL, 12),
    Elem(12): (VAL, 13),
    Elem(13): (VAL, 14),
    Elem(14): (VAL, 15),
    Elem(15): (VAL, 16),
    Elem(16): (VAL, 17),
    Elem(17): (VAL, 18),
}

PREINFUSION_MAP = {
    Elem(CALCULATED_VALUE): ENABLE_PREINFUSION,
    Elem(0): (PREINFUSION, "k1"),
    Elem(1): (PREINFUSION, "k2"),
    Elem(2): (PREINFUSION, "k3"),
    Elem(3): (PREINFUSION, "k4"),
}

# 64
# 00
# 64
# 00
# 70
# 00
# 00
# 00
# 00
# 01
# 01
# 69: Group Offset (0-200 C, 0-360 F, +/- from the midpoint)
# 00
# 55
# 00
# 00

FACTORY_CONFIG_MAP = {
    Elem(3): T_UNIT,
    Elem(20): PID_OFFSET,
    Elem(22): WATER_FILTER_LITERS,
    Elem(27, 2): BREW_GROUP_OFFSET,
    Elem(29): FACTORY_OFFSET,
}

class Msg:
    GET_STATUS = "get_status"
    GET_SHORT_STATUS = "get_short_status"
    GET_CONFIG = "get_config"
    GET_AUTO_ON_OFF_TIMES = "get_auto_sched"
    SET_AUTO_ON_OFF_TIMES = "set_auto_on_off_times"
    SET_POWER = "set_power"
    GET_TEMP_REPORT = "get_temp_report"
    GET_AUTO_ON_OFF_ENABLE = "get_auto_on_off_enable"
    SET_AUTO_ON_OFF_ENABLE = "set_auto_on_off_enable"
    GET_SER_NUM = "get_ser_num"
    GET_DATETIME = "get_datetime"
    SET_COFFEE_TEMP = "set_coffee_temp"
    SET_STEAM_TEMP = "set_steam_temp"
    SET_PREBREWING_ENABLE = "set_prebrewing_enable"
    SET_DOSE = "set_dose"
    SET_DOSE_HOT_WATER = "set_dose_hot_water"
    SET_PREBREW_TIMES = "set_prebrew_times"
    GET_DRINK_STATS = "get_drink_stats"
    GET_WATER_FLOW = "get_water_flow"
    GET_USAGE_STATS = "get_usage_stats"
    GET_FRONT_DISPLAY = "get_front_display"
    GET_MYSTERY = "get_mystery"
    SET_START_BACKFLUSH = "set_start_backflush"
    GET_STATUS_MYSTERY = "get_status_mystery"
    GET_PREINFUSION_TIMES = "get_preinfusion_times"
    SET_PREINFUSION_TIME = "set_preinfusion_time"
    SET_STEAM_BOILER_ENABLE = "set_steam_boiler_enable"
    GET_FACTORY_CONFIG = "get_factory_config"

    """Auto on/off address starts at 0x11 for Monday and increments by 4 thereafter."""
    AUTO_ON_OFF_HOUR_BASE = 0x11
    AUTO_ON_OFF_MIN_BASE = 0x20

    """Dose address starts at 0x14 for the first key and increments by 2 thereafter."""
    DOSE_KEY_BASE = 0x14

    """Prebrew on address base."""
    PREBREW_ON_BASE = 0x0C

    """Prebrew off address base."""
    PREBREW_OFF_BASE = 0x10

    """Preinfusion address base."""
    PREINFUSION_BASE = 0xE2

    READ = "R"
    WRITE = "W"
    STREAM = "Z"

    RESPONSE_GOOD = "OK"

    def __init__(self, msg_type, msg, map):
        """Init command"""
        self._msg_type = msg_type
        self._msg = msg
        self._map = map

    @property
    def msg_type(self):
        """Command type"""
        return self._msg_type

    @property
    def msg(self):
        """Command type"""
        return self._msg

    @property
    def map(self):
        """Return map if we should decode, None if not"""
        return self._map


MSGS = {
    # Reads
    Msg.GET_STATUS: Msg(Msg.READ, "40000023", STATUS_MAP),
    Msg.GET_SHORT_STATUS: Msg(Msg.READ, "40000020", None),
    Msg.GET_CONFIG: Msg(Msg.READ, "0000001F", CONFIG_MAP),
    Msg.GET_AUTO_ON_OFF_TIMES: Msg(Msg.READ, "0310001D", AUTO_SCHED_MAP),
    Msg.GET_AUTO_ON_OFF_ENABLE: Msg(Msg.READ, "03100001", AUTO_ENABLE_MAP),
    Msg.GET_TEMP_REPORT: Msg(Msg.READ, "401C0004", TEMP_REPORT_MAP),
    Msg.GET_SER_NUM: Msg(Msg.READ, "01000011", SER_NUM_MAP),
    Msg.GET_DATETIME: Msg(Msg.READ, "03000007", DATETIME_MAP),
    Msg.GET_DRINK_STATS: Msg(Msg.READ, "0020002C", DRINK_STATS_MAP),
    Msg.GET_WATER_FLOW: Msg(Msg.STREAM, "60000016", WATER_FLOW_MAP),
    Msg.GET_USAGE_STATS: Msg(Msg.READ, "00500018", USAGE_MAP),
    Msg.GET_FRONT_DISPLAY: Msg(Msg.READ, "60EF0010", FRONT_DISPLAY_MAP),
    Msg.GET_MYSTERY: Msg(Msg.READ, "60DA0012", MYSTERY_MAP),
    Msg.GET_STATUS_MYSTERY: Msg(Msg.READ, "40220001", None),
    Msg.GET_PREINFUSION_TIMES: Msg(Msg.READ, "00E20004", PREINFUSION_MAP),
    Msg.GET_FACTORY_CONFIG: Msg(Msg.READ, "40600020", FACTORY_CONFIG_MAP),
    # Writes
    Msg.SET_POWER: Msg(Msg.WRITE, "00000001", None),
    Msg.SET_COFFEE_TEMP: Msg(Msg.WRITE, "00070002", None),
    Msg.SET_STEAM_TEMP: Msg(Msg.WRITE, "00090002", None),
    Msg.SET_PREBREWING_ENABLE: Msg(Msg.WRITE, "000B0001", None),
    Msg.SET_AUTO_ON_OFF_TIMES: Msg(Msg.WRITE, "03100002", None),
    Msg.SET_AUTO_ON_OFF_ENABLE: Msg(Msg.WRITE, "03100001", None),
    # Write config for keys (second byte is base + the key number and will be replaced)
    Msg.SET_DOSE: Msg(Msg.WRITE, "00140002", None),
    Msg.SET_DOSE_HOT_WATER: Msg(Msg.WRITE, "001E0001", None),
    Msg.SET_PREBREW_TIMES: Msg(Msg.WRITE, "000C0001", None),
    Msg.SET_START_BACKFLUSH: Msg(Msg.WRITE, "00E10001", None),
    Msg.SET_PREINFUSION_TIME: Msg(Msg.WRITE, "00E20001", None),
    Msg.SET_STEAM_BOILER_ENABLE: Msg(Msg.WRITE, "00E10001", None),
}

# 0000:001f - Config
# 0020:004C - Drink Stats
# 0050:0068 - Usage Stats
# 0070:00AF - Unused (FF)
# 00B0:00BB - Module Serial Number
# 00C0:00CB - Module Serial Number
# 00D0:00DF - Machine Name
# 00E0:00FF - Unused (00)
# 0100:010F - Machine Serial Number
# 0110:011F -
